'''
Created with love by Sigmoid
@Author - Basoc Nicoleta-Nina - nicoleta.basoc28@gmail.com
'''

import pandas as pd
import numpy as np
from sklearn.neighbors import NearestNeighbors
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError


class MultiSURFSelector:

    def __init__(self, n_neighbours: int = 7, n_features: int = 5) -> None:
        """
            Setting up the algorithm
        :param n_neighbors: int, default = 7
            The number of neighbors to consider when assigning feature
            importance scores. If a float number is provided, that percentage of
            training samples is used as the number of neighbors.
            More neighbors results in more accurate scores, but takes longer.
            
        :param n_features: int, default = 5
            The number of top features (according to the relieff score) to
            retain after feature selection is applied.
        """
        
        self.n_neighbours = n_neighbours
        self.n_features = n_features
        
    def define_tresh_multisurf(self, dataframe : pd.DataFrame, target : str) -> float:
        """
            Defining the treshold for the multiSURF algorithm
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: str
             The column name of the value that we what to predict
        :return: float
            
        """
        
        X=dataframe.drop([target], axis=1).values
        distance_vector=[]
        # Defining the distance array based on which we will calculate the threshold.
        for i in range(len(X)):
            for j in range(len(X)):
                if i!=j:
                    distance_vector.append(np.linalg.norm(X[i]-X[j]))
        
        # Calculating the average distances and their stds and the value of the threshold.
        average_distance=np.array(distance_vector).mean()
        std_vector=np.std(distance_vector)/2
        diference_tresh_std=average_distance - std_vector
        return diference_tresh_std
        
    def select(self, dataframe : pd.DataFrame, target : str) -> list:
        """
            Selecting the most important columns
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param target: str
             The column name of the value that we what to predict
        :return: list
            The list of features that our algorithm selects as the best ones
        """
        # Verifying if data frame has NaN-values.
        if dataframe.isna().values.any():
            raise MissingDataError("The passed data frame contains missing values")
            
        # if the given target doesn't match any column in the DataFrame raise an error
        if target not in dataframe:
            raise NoSuchColumnError(f'No such column {target}')
            
       # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError('The given DataFrame contains non-numeric values')
            
        # Converting the data frame in numpy arrays.
        X = dataframe.drop([target], axis=1).values
        y = dataframe[target].values
        
        # Obtaining the number of features and samples.
        n_samples, feature_number = X.shape
        
        # Initialising the score of all features as 0.
        score = np.zeros(feature_number)
        
        # Using NearestNeighbors to define the nighbours for our target instance and all others.
        nbrs = NearestNeighbors(n_neighbors=n_samples,algorithm='kd_tree').fit(X)
        
        # Getting the distances of the neighbours and their indices.
        distances, indices = nbrs.kneighbors(X)
        
        # Getting how many classes we have to predict. 
        n_classes = np.unique(y).tolist()
        
        self.treshold=self.define_tresh_multisurf(dataframe, target)
        
        # Identifying hits and misses.
        for i in range(n_samples):
            
            # Initialising the miss and hit counters. 
            miss=dict()
            hit=[]
            
            # Getting all the classes of the target column.
            for j in n_classes:
                miss[j] = []
            
            # Finding misses and hits based on the threshold. 
            for n in range(1, self.n_neighbors):
                if np.linalg.norm(y[i] - y[indices[i][n]])<self.treshold:
                    if y[i] == y[indices[i][n]]:
                        hit.append(indices[i][n])
                    else:
                        miss[y[indices[i][n]]].append(indices[i][n])
            
            # Feature Weight update based on the right misses and hits.
            cont = dict()
            cont_diff_idx = float(len(y[y == y[i]])) / float(n_samples)
            
            # Computing the probabilities of every class.
            for label in n_classes:
                cont_diff_class = float(len(y[y == label])) / float(n_samples)
                cont[label] = cont_diff_class/ (1 - cont_diff_idx)
                
            # For hits.
            hit_term = np.zeros(feature_number)
            for h in hit:
                hit_term = np.array(hit_term) + np.array(abs(X[i, :] - X[h, :]))
                
            # For misses.
            miss_term = dict()
            for (label, miss_list) in miss.items():
                miss_term[label] = np.zeros(feature_number)
                for m in miss_list:
                    miss_term[label] = np.array(abs(X[i, :] - X[m, :])) + np.array(miss_term[label])
                
                # Update score for miss.
                score += miss_term[label] / (self.n_neighbors * n_samples)
                
            # Update score for hit.
            score -= hit_term / self.n_neighbors

        # Sorting the indexes based on the score.    
        result = np.argsort(score, 0)
        index = result[::-1][:self.n_features]
        
        # Return the columns based on their score.
        return list(dataframe.iloc[:, index].columns)                           



'''
Created with love by Sigmoid

@Author - Păpăluță Vasile - vpapaluta06@gmail.com
'''
# Importing all needed libraries
from .errors import NotBetweenZeroAndOneError

class PearsonCorrelationSelector:
    def __init__(self, min_corr : float = 0.5, max_corr : float = 0.8, erase_corr : bool = False) -> None:
        '''
            Setting up the algorithm
        :param min_corr: float, between 0 and 1, default = 0.5
            The minimal positive correlation value that must the feature have with y_column
        :param max_corr: float, between 0 and 1, default = 0.8
            The maximal positive correlation value that must the feature have with y_column
        :param erase_corr: boolean, default = False
            If set as False the selector doesn't erase features that are highly correlated between themselves
            If set as True the selector does erase features that are highly correlated between themselves
        '''
        try:
            if min_corr < 0 or max_corr > 1:
                raise NotBetweenZeroAndOneError
            if str(type(erase_corr)) != "<class 'bool'>":
                raise ValueError
        except NotBetweenZeroAndOneError:
            print("Min or Max Correlations are not seted between 0 and 1!")
            quit()
        except ValueError:
            print("Parameter erase_col isn't set as a boolean value!")
            quit()
        finally:
            self.min_corr = min_corr
            self.max_corr = max_corr
            self.erase_corr = erase_corr
    def index_to_cols(self, index_list : list) -> list:
        '''
            Converting the list of indexes in a list of features that will be picked by the model
        :param index_list: list
            A list of Indexes that should be converted into a list of feature names
        :return: list
            A list with feature names
        '''
        return [self.X_columns[i] for i in index_list]
    def select(self, dataframe : 'pd.Dataframe', y_column : str) -> list:
        '''
            Selecting the most important features
        :param dataframe: pandas DataFrame
             Data Frame on which the algorithm is applied
        :param y_column: string
             The column name of the value that we what to predict
        :return: list
            The list of features that are selected by the algorithm as the best one
        '''
        # Getting the list with names of columns without the target one
        self.X_columns = [col for col in dataframe.columns if col != y_column]

        # Creating an empty list for correlated columns
        correlated_indexes = []

        # Getting the y-column index
        y_column_index = list(dataframe.columns).index(y_column)

        # Generating the correlation matrix
        self.corr_table = dataframe.corr()
        corr_matrix = self.corr_table.values

        # Searching for the columns correlated with the y-column
        for i in range(len(corr_matrix[y_column_index])):
            if abs(corr_matrix[y_column_index][i]) > self.min_corr and abs(corr_matrix[y_column_index][i]) < self.max_corr:
                correlated_indexes.append(i)

        # Creating a list with the names of columns correlated with the y-column
        self.correlated_cols = self.index_to_cols(correlated_indexes)

        # If we chose to erase a column from the correlated pairs we erase one of them
        if self.erase_corr:
            cols_to_remove = []
            for i in correlated_indexes:
                for j in correlated_indexes:
                    if abs(corr_matrix[i][j]) > self.min_corr and abs(corr_matrix[i][j]) < self.max_corr:
                        cols_to_remove.append(self.X_columns[i])
            cols_to_remove = set(cols_to_remove)

            # Removing the chose columns
            for col in cols_to_remove:
                self.correlated_cols.remove(col)
        return self.correlated_cols




from setuptools import setup
long_description = '''
# kydavra
Kydavra is a python sci-kit learn inspired package for feature selection. It used some statistical methods to extract from pure pandas Data Frames the columns that are related to column that your model should predict.
This version of kydavra has the next methods of feature selection:
* ANOVA test selector (ANOVASelector)
* Chi squared selector (ChiSquaredSelector)
* Genetic Algorithm selector (GeneticAlgorithmSelector)
* Kendall Correlation selector (KendallCorrelationSelector)
* Lasso selector (LassoSelector)
* Pearson Correlation selector (PearsonCorrelationSelector)
* Point-Biserial selector (PointBiserialCorrSelector)
* P-value selector (PValueSelector)
* Spearman Correlation selector (SpermanCorrelationSelector)
* Shannon selector (ShannonSelector)
* ElasticNet Selector (ElasticNetSelector)
* M3U Selector (M3USelector)
* MUSE Selector (MUSESelector)
* Mixer Selector (MixerSelector)
* PCA Filter (PCAFilter)
* PCA Reducer (PCAReducer)
* LDA Reducer (LDAReducer)
* Bregman Divergence selector (BregmanDivergenceSelector)
* Fisher Selector (FisherSelector)
* ICA Reducer (ICAReducer)
* ICA Filter (ICAFilter)
* Itakura-Saito Divergence selector (ItakuraSaitoSelector)
* Jensen-Shannon Divergence selector (JensenShannonSelector)
* Kullback-Leibler selector (KullbackLeiblerSelector)
* MultiSURF selector (MultiSURFSelector)
* Phik selector (PhikSelector)
* ReliefF selector (ReliefFSelector)

All these methods takes the pandas Data Frame and y column to select from remained columns in the Data Frame.

How to use kydavra\
To use selector from kydavra you should just import the selector from kydavra in the following framework:
```python
from kydavra import PValueSelector
```
class names are written above in parantheses.\
Next create a object of this algorithm (I will use p-value method as an example).
```python
method = PValueSelector()
```
To get the best feature on the opinion of the method you should use the 'select' function, using as parameters the pandas Data Frame and the column that you want your model to predict.
```python
selected_columns = method.select(df, 'target')
```
Returned value is a list of columns selected by the algorithm.

Some methods could plot the process of selecting the best features.\
In these methods dotted are features that wasn't selected by the method.\
*ChiSquaredSelector*
```python
method.plot_chi2()
```
For ploting and
```python
method.plot_chi2(save=True, file_path='FILE/PATH.png')
```
and
```python
method.plot_p_value()
```
for ploting the p-values.\
*LassoSelector*
```python
method.plot_process()
```
also you can save the plot using the same parameters.\
*PValueSelector*
```
method.plot_process()
```

Some advice.
* Use ChiSquaredSelector for categorical features.
* Use LassoSelector and PValueSelector for regression problems.
* Use PointBiserialCorrSelector for binary classification problems.
* Use ShannonSelector to choose whatever to keep the NaN values (as another value) and to drop column with a lot of NaN values.\n

With love from Sigmoid.

We are open for feedback. Please send your impression to vladimir.stojoc@gmail.com
'''
setup(
  name = 'kydavra',
  packages = ['kydavra'],
  version = '0.3.4',
  license='MIT',
  description = 'Kydavra is a sci-kit learn inspired python library with feature selection methods for Data Science and Macine Learning Model development',
  long_description=long_description,
  long_description_content_type='text/markdown',
  author = 'SigmoidAI - Păpăluță Vasile, Stojoc Vladimir',
  author_email = 'vladimir.stojoc@gmail.com',
  url = 'https://github.com/SigmoidAI/kydavra',
  download_url = 'https://github.com/ScienceKot/kydavra/archive/v1.0.tar.gz',    # I explain this later on
  keywords = ['ml', 'machine learning', 'feature selection', 'python'],   # Keywords that define your package best
  install_requires=[            # I get to this in a second
          'numpy',
          'pandas',
          'scikit-learn',
          'statsmodels',
          'matplotlib',
          'seaborn'
      ],
  classifiers=[
    'Development Status :: 3 - Alpha',
    'Framework :: Jupyter',
    'Intended Audience :: Science/Research',
    'Topic :: Scientific/Engineering :: Artificial Intelligence',
    'Topic :: Scientific/Engineering :: Mathematics',
    'Topic :: Software Development :: Libraries :: Python Modules',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3.4',
    'Programming Language :: Python :: 3.5',
    'Programming Language :: Python :: 3.6',
    'Programming Language :: Python :: 3.7',
    'Programming Language :: Python :: 3.8',
  ],
)


'''
Created with love by Sigmoid
@Author - iurie.cius.personal@gmail.com
'''

# Importing all needed libraries.
import numpy as np
import pandas as pd
from sklearn.neighbors import NearestNeighbors
from .errors import MissingDataError, NonNumericDataError, NoSuchColumnError


class ReliefFSelector:
    def __init__(self, n_neighbors: int = 5, n_features: int = 10) -> None:
        """
            Setting up the algorithm
        :param n_neighbors: int, default = 5
            The number of neighbors to consider when assigning feature
            importance scores. If a float number is provided, that percentage of
            training samples is used as the number of neighbors.
            More neighbors results in more accurate scores, but takes longer.
        :param n_features: int, default = 10
            The number of top features (according to the relieff score) to
            retain after feature selection is applied.
        """
        self.n_neighbors = n_neighbors
        self.n_features = n_features

    def select(self, dataframe: 'pd.DataFrame', target: str) -> list:
        """
            Selecting the most important columns.
        :param dataframe: pandas DataFrame
            Data Frame on which the algorithm is applied.
        :param target: str
            The column name of the value that we what to predict.
        :return: list
            The list of features that are selected by the algorithm as the best one.
        """

        if dataframe.isna().values.any():
            raise MissingDataError(
                "The passed data frame contains missing values.")

        # if a column has string values raise an error
        for col in dataframe:
            if not pd.api.types.is_numeric_dtype(dataframe[col]):
                raise NonNumericDataError(
                    'The given DataFrame contains non-numeric values !')

        if target not in dataframe:
            raise NoSuchColumnError("The passed target column doesn't exist.")

        # Converting the data frame in numpy arrays.
        X = dataframe.drop([target], axis=1).values
        y = dataframe[target].values

        # Get the dimensionality of the DataFrame.
        n_samples, n_features = X.shape

        # Initialize the scores to 0.
        score = np.zeros(n_features)

        # Identify neighbors.
        nbrs = NearestNeighbors(n_neighbors=n_samples,
                                algorithm='kd_tree').fit(X)

        # Get the number of classes.
        c = np.unique(y).tolist()

        # Compute the distances and indices of the neighbors.
        distances, indices = nbrs.kneighbors(X)

        # Iterating through every sample in the dataset.
        for i in range(n_samples):

            # Initialize the near miss and near hit data structures.
            near_miss = dict()
            near_hit = []

            # Getting all the classes of the target column.
            for label in c:
                near_miss[label] = []

            # Identify k nearest hits and k nearest misses (using distance array).
            for n in range(1, self.n_neighbors):
                if y[i] == y[indices[i][n]]:
                    near_hit.append(indices[i][n])
                else:
                    near_miss[y[indices[i][n]]].append(indices[i][n])

            # Initialize the dictionary for storing the probabilities of every class.
            p_dict = dict()
            p_label_index = float(len(y[y == y[i]])) / float(n_samples)

            # Computing the probabilities of every class.
            for label in c:
                p_label_c = float(len(y[y == label])) / float(n_samples)
                p_dict[label] = p_label_c / (1 - p_label_index)

            # Initialize near hit terms.
            near_hit_term = np.zeros(n_features)

            # Compute near hit terms.
            for element in near_hit:
                near_hit_term = np.array(
                    near_hit_term) + np.array(abs(X[i, :] - X[element, :]))

            # Initialize near miss terms.
            near_miss_term = dict()

            # Compute near miss terms.
            for (label, miss_list) in near_miss.items():
                near_miss_term[label] = np.zeros(n_features)
                for element in miss_list:
                    near_miss_term[label] = np.array(abs(
                        X[i, :] - X[element, :])) + np.array(near_miss_term[label])

                # Update the Score.
                score += near_miss_term[label] / (self.n_neighbors * n_samples)

            # Update the Score.
            score -= near_hit_term / self.n_neighbors

        # Sorting the indexes based on the score.
        res = np.argsort(score, 0)
        idx = res[::-1][:self.n_features]

        # Return the top n features.
        return list(dataframe.iloc[:, idx].columns)
